from __future__ import annotations

from icclim.ecad.xclim_binding import XCLIM_BINDING
from icclim.generic_indices.generic_indicators import GenericIndicatorRegistry
from icclim.generic_indices.standard_variable import StandardVariableRegistry
from icclim.generic_indices.threshold import build_threshold
from icclim.models.constants import ECAD_ATBD, QUANTILE_BASED, REFERENCE_PERIOD_INDEX
from icclim.models.index_group import IndexGroupRegistry
from icclim.models.registry import Registry
from icclim.models.standard_index import StandardIndex

ECAD_REFERENCE = (
    "ATBD of the ECA&D indices calculation"
    " (https://knmi-ecad-assets-prd.s3.amazonaws.com/documents/atbd.pdf)"
)


class EcadIndexRegistry(Registry[StandardIndex]):
    _item_class = StandardIndex
    # TODO Add indices of wind gust, wind direction,
    #                     radiation, pressure,
    #                     cloud cover, sunshine,
    #                     humidity

    @staticmethod
    def get_item_aliases(item: StandardIndex) -> list[str]:
        return [item.short_name]

    @classmethod
    def list(cls: EcadIndexRegistry) -> list[str]:
        return [
            f"{i.group.name} | {i.short_name} | {i.definition}" for i in cls.values()
        ]

    TG = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        output_unit="degree_Celsius",
        definition="Mean of daily mean temperature",
        source=ECAD_ATBD,
        short_name="TG",
        group=IndexGroupRegistry.TEMPERATURE,
        input_variables=[StandardVariableRegistry.TAS],
    )
    TN = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        output_unit="degree_Celsius",
        definition="Mean of daily minimum temperature",
        source=ECAD_ATBD,
        short_name="TN",
        group=IndexGroupRegistry.TEMPERATURE,
        input_variables=[StandardVariableRegistry.TAS_MIN],
    )
    TX = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        output_unit="degree_Celsius",
        definition="Mean of daily maximum temperature",
        source=ECAD_ATBD,
        short_name="TX",
        group=IndexGroupRegistry.TEMPERATURE,
        input_variables=[StandardVariableRegistry.TAS_MAX],
    )
    DTR = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.MeanOfDifference,
        output_unit="degree_Celsius",
        definition="Mean Diurnal Temperature Range",
        source=ECAD_ATBD,
        short_name="DTR",
        group=IndexGroupRegistry.TEMPERATURE,
        input_variables=[
            StandardVariableRegistry.TAS_MAX,
            StandardVariableRegistry.TAS_MIN,
        ],
    )
    ETR = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.DifferenceOfExtremes,
        output_unit="degree_Celsius",
        definition="Intra-period extreme temperature range",
        source=ECAD_ATBD,
        short_name="ETR",
        group=IndexGroupRegistry.TEMPERATURE,
        input_variables=[
            StandardVariableRegistry.TAS_MAX,
            StandardVariableRegistry.TAS_MIN,
        ],
    )
    VDTR = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.MeanOfAbsoluteOneTimeStepDifference,
        output_unit="degree_Celsius",
        definition="Mean day-to-day variation in Diurnal Temperature Range",
        source=ECAD_ATBD,
        short_name="vDTR",
        group=IndexGroupRegistry.TEMPERATURE,
        input_variables=[
            StandardVariableRegistry.TAS_MAX,
            StandardVariableRegistry.TAS_MIN,
        ],
    )
    # Heat
    SU = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        definition="Number of Summer Days (Tmax > 25C)",
        source=ECAD_ATBD,
        short_name="SU",
        threshold="> 25 degree_Celsius",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[],
    )
    TR = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        definition="Number of Tropical Nights (Tmin > 20C)",
        source=ECAD_ATBD,
        short_name="TR",
        threshold="> 20 degree_Celsius",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[],
    )
    WSDI = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.SumOfSpellLengths,
        output_unit="day",
        definition="Warm-spell duration index (days)",
        source=ECAD_ATBD,
        short_name="WSDI",
        threshold="> 90 doy_per",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
        min_spell_length=6,
    )
    TG90P = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        definition="Days when Tmean > 90th percentile",
        threshold="> 90 doy_per",
        source=ECAD_ATBD,
        short_name="TG90p",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TN90P = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        definition="Days when Tmin > 90th percentile",
        threshold="> 90 doy_per",
        source=ECAD_ATBD,
        short_name="TN90p",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TX90P = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="> 90 doy_per",
        output_unit="day",
        definition="Days when Tmax > 90th daily percentile",
        source=ECAD_ATBD,
        short_name="TX90p",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TXX = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Maximum,
        output_unit="degree_Celsius",
        definition="Maximum daily maximum temperature",
        source=ECAD_ATBD,
        short_name="TXx",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
    )
    TNX = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Minimum,
        output_unit="degree_Celsius",
        definition="Maximum daily minimum temperature",
        source=ECAD_ATBD,
        short_name="TNx",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MIN],
    )
    CSU = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.MaxConsecutiveOccurrence,
        threshold="> 25 degree_Celsius",
        output_unit="day",
        definition="Maximum number of consecutive summer days (Tmax >25 C)",
        source=ECAD_ATBD,
        short_name="CSU",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[],
    )
    # Cold
    GD4 = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Excess,
        threshold="4 degree_Celsius",
        output_unit="degree_Celsius day",
        definition="Growing degree days (sum of Tmean > 4 C)",
        source=ECAD_ATBD,
        short_name="GD4",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS],
        qualifiers=[],
    )
    FD = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="< 0 degree_Celsius",
        output_unit="day",
        definition="Number of Frost Days (Tmin < 0C)",
        source=ECAD_ATBD,
        short_name="FD",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[],
    )
    CFD = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.MaxConsecutiveOccurrence,
        threshold="< 0 degree_Celsius",
        output_unit="day",
        definition="Maximum number of consecutive frost days (Tmin < 0 C)",
        source=ECAD_ATBD,
        short_name="CFD",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[],
    )
    HD17 = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Deficit,
        threshold="17 degree_Celsius",
        output_unit="degree_Celsius day",
        definition="Heating degree days (sum of Tmean < 17 C)",
        source=ECAD_ATBD,
        short_name="HD17",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS],
        qualifiers=[],
    )
    ID = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="< 0 degree_Celsius",
        output_unit="day",
        definition="Number of sharp Ice Days (Tmax < 0C)",
        source=ECAD_ATBD,
        short_name="ID",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[],
    )
    TG10P = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="< 10 doy_per",
        output_unit="day",
        definition="Days when Tmean < 10th percentile",
        source=ECAD_ATBD,
        short_name="TG10p",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TN10P = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="< 10 doy_per",
        output_unit="day",
        definition="Days when Tmin < 10th percentile",
        source=ECAD_ATBD,
        short_name="TN10p",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TX10P = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="< 10 doy_per",
        output_unit="day",
        definition="Days when Tmax < 10th percentile",
        source=ECAD_ATBD,
        short_name="TX10p",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TXN = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Minimum,
        output_unit="degree_Celsius",
        definition="Minimum daily maximum temperature",
        source=ECAD_ATBD,
        short_name="TXn",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MAX],
    )
    TNN = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Minimum,
        output_unit="degree_Celsius",
        definition="Minimum daily minimum temperature",
        source=ECAD_ATBD,
        short_name="TNn",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MIN],
    )
    CSDI = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.SumOfSpellLengths,
        threshold="< 10 doy_per",
        output_unit="day",
        definition="Cold-spell duration index (days)",
        source=ECAD_ATBD,
        short_name="CSDI",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
        min_spell_length=6,
    )
    # Drought
    CDD = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.MaxConsecutiveOccurrence,
        threshold="< 1 mm/day",
        output_unit="day",
        definition="Maximum consecutive dry days (Precip < 1mm)",
        source=ECAD_ATBD,
        short_name="CDD",
        group=IndexGroupRegistry.DROUGHT,
        input_variables=[StandardVariableRegistry.PR],
    )
    # Rain
    PRCPTOT = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Sum,
        threshold=">= 1 mm/day",
        output_unit="mm",
        definition="Total precipitation during Wet Days",
        source=ECAD_ATBD,
        short_name="PRCPTOT",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    RR1 = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=">= 1 mm/day",
        output_unit="day",
        definition="Number of Wet Days (precip >= 1 mm)",
        source=ECAD_ATBD,
        short_name="RR1",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    SDII = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        threshold=">= 1 mm/day",
        output_unit="mm/day",
        definition="Average precipitation during Wet Days (SDII)",
        source=ECAD_ATBD,
        short_name="SDII",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    CWD = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.MaxConsecutiveOccurrence,
        threshold=">= 1 mm/day",
        output_unit="day",
        definition="Maximum consecutive wet days (Precip >= 1mm)",
        source=ECAD_ATBD,
        short_name="CWD",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    RR = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Sum,
        output_unit="mm",
        definition="Precipitation sum (mm)",
        source=ECAD_ATBD,
        short_name="RR",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    R10MM = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=">= 10 mm/day",
        output_unit="day",
        definition="Number of heavy precipitation days (Precip >=10mm)",
        source=ECAD_ATBD,
        short_name="R10mm",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    R20MM = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=">= 20 mm/day",
        output_unit="day",
        definition="Number of very heavy precipitation days (Precip >= 20mm)",
        source=ECAD_ATBD,
        short_name="R20mm",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    RX1DAY = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Maximum,
        output_unit="mm/day",
        definition="maximum 1-day total precipitation",  # from xclim
        source=ECAD_ATBD,
        short_name="RX1day",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    RX5DAY = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.MaxOfRollingSum,
        output_unit="mm",
        definition="maximum 5-day total precipitation",  # from xclim
        source=ECAD_ATBD,
        short_name="RX5day",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[],
        rolling_window_width=5,
    )
    R75P = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=build_threshold("> 75 period_per", threshold_min_value="1 mm/day"),
        output_unit="day",
        definition="Days with RR > 75th percentile of daily amounts (moderate wet days)"
        " (d)",
        source=ECAD_ATBD,
        short_name="R75p",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    R75PTOT = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.FractionOfTotal,
        threshold=build_threshold("> 75 period_per", threshold_min_value="1 mm/day"),
        output_unit="%",
        definition="Precipitation fraction due to moderate wet days"
        " (> 75th percentile)",
        source=ECAD_ATBD,
        short_name="R75pTOT",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    R95P = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=build_threshold("> 95 period_per", threshold_min_value="1 mm/day"),
        output_unit="day",
        definition="Days with RR > 95th percentile of daily amounts (very wet days)"
        " (days)",
        source=ECAD_ATBD,
        short_name="R95p",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    R95PTOT = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.FractionOfTotal,
        threshold=build_threshold("> 95 period_per", threshold_min_value="1 mm/day"),
        output_unit="%",
        definition="Precipitation fraction due to very wet days (> 95th percentile)",
        source=ECAD_ATBD,
        short_name="R95pTOT",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    R99P = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=build_threshold("> 99 period_per", threshold_min_value="1 mm/day"),
        output_unit="day",
        definition="Days with RR > 99th percentile of daily amounts"
        " (extremely wet days)",
        source=ECAD_ATBD,
        short_name="R99p",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    R99PTOT = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.FractionOfTotal,
        threshold=build_threshold("> 99 period_per", threshold_min_value="1 mm/day"),
        output_unit="%",
        definition="Precipitation fraction due to extremely wet days"
        " (> 99th percentile)",
        source=ECAD_ATBD,
        short_name="R99pTOT",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    # Snow
    SD = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        output_unit="cm",
        definition="Mean of daily snow depth",
        source=ECAD_ATBD,
        short_name="SD",
        group=IndexGroupRegistry.SNOW,
        input_variables=[StandardVariableRegistry.SND],
    )
    SD1 = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=">= 1 cm",
        output_unit="day",
        definition="Snow days (SD >= 1 cm)",
        source=ECAD_ATBD,
        short_name="SD1",
        group=IndexGroupRegistry.SNOW,
        input_variables=[StandardVariableRegistry.SND],
    )
    SD5CM = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        threshold=">= 5 cm",
        definition="Number of days with snow depth >= 5 cm",
        source=ECAD_ATBD,
        short_name="SD5cm",
        group=IndexGroupRegistry.SNOW,
        input_variables=[StandardVariableRegistry.SND],
    )
    SD50CM = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=">= 50 cm",
        output_unit="day",
        definition="Number of days with snow depth >= 50 cm",
        source=ECAD_ATBD,
        short_name="SD50cm",
        group=IndexGroupRegistry.SNOW,
        input_variables=[StandardVariableRegistry.SND],
    )
    # Compound (precipitation and temperature)
    CD = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=[
            "< 25 doy_per",
            build_threshold("< 25 period_per", threshold_min_value="1 mm/day"),
        ],
        output_unit="day",
        definition="Days with TG < 25th percentile of daily mean temperature and"
        " RR <25th percentile of daily precipitation sum (cold/dry days)",
        source=ECAD_ATBD,
        short_name="CD",
        group=IndexGroupRegistry.TEMPERATURE | IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.TAS, StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    CW = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=[
            "< 25 doy_per",
            build_threshold("> 75 period_per", threshold_min_value="1 mm/day"),
        ],
        output_unit="day",
        definition="Days with TG < 25th percentile of daily mean temperature and"
        " RR >75th percentile of daily precipitation sum (cold/wet days)",
        source=ECAD_ATBD,
        short_name="CW",
        group=IndexGroupRegistry.TEMPERATURE | IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.TAS, StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    WD = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=[
            "> 75 doy_per",
            build_threshold("< 25 period_per", threshold_min_value="1 mm/day"),
        ],
        output_unit="day",
        definition="Days with TG > 75th percentile of daily mean temperature and"
        " RR <25th percentile of daily precipitation sum (warm/dry days)",
        source=ECAD_ATBD,
        short_name="WD",
        group=IndexGroupRegistry.TEMPERATURE | IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.TAS, StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    WW = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=[
            "> 75 doy_per",
            build_threshold("> 75 period_per", threshold_min_value="1 mm/day"),
        ],
        output_unit="day",
        definition="Days with TG > 75th percentile of daily mean temperature and"
        " RR >75th percentile of daily precipitation sum (warm/wet days)",
        source=ECAD_ATBD,
        short_name="WW",
        group=IndexGroupRegistry.TEMPERATURE | IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.TAS, StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    # WIND
    FXX = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Maximum,
        output_unit="m s-1",
        definition="Maximum value of daily maximum wind gust",
        source=ECAD_ATBD,
        short_name="FXx",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.WSGS_MAX],
    )
    FG6BFT = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=">= 10.8 m s-1",
        output_unit="day",
        definition="Days with daily averaged wind ≥ 6 Bft (10.8 m s-1)",
        source=ECAD_ATBD,
        short_name="FG6Bft",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.SFC_WIND],
    )
    FGCALM = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="<= 2 m s-1",
        output_unit="day",
        definition="Calm days, days with daily averaged wind <= 2 m s-1",
        source=ECAD_ATBD,
        short_name="FGcalm",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.SFC_WIND],
    )
    FG = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        output_unit="m s-1",
        definition="Mean of daily mean wind strength",
        source=ECAD_ATBD,
        short_name="FG",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.SFC_WIND],
    )
    DDNORTH = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="> 315 degree OR <= 45 degree",
        output_unit="day",
        definition="Days with northerly winds (DD > 315° or DD ≤ 45°)",
        source=ECAD_ATBD,
        short_name="DDnorth",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.WIND_TO_DIRECTION],
    )
    DDEAST = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="> 45 degree AND <= 135 degree",
        output_unit="day",
        definition="Days with easterly winds (45° < DD <= 135°)",
        source=ECAD_ATBD,
        short_name="DDeast",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.WIND_TO_DIRECTION],
    )
    DDSOUTH = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="> 135 degree AND <= 225 degree",
        output_unit="day",
        definition="Days with southerly winds (135° < DD <= 225°)",
        source=ECAD_ATBD,
        short_name="DDsouth",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.WIND_TO_DIRECTION],
    )
    DDWEST = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="> 225 degree AND <= 315 degree",
        output_unit="day",
        definition="Days with westerly winds (225° < DD <= 315°)",
        source=ECAD_ATBD,
        short_name="DDwest",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.WIND_TO_DIRECTION],
    )
    GSL = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=XCLIM_BINDING.GrowingSeasonLength(),
        definition="Growing season length",
        source=ECAD_ATBD,
        short_name="GSL",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS],
        output_unit="day",
    )
    SPI6 = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=XCLIM_BINDING.StandardizedPrecipitationIndex6(),
        definition="6-Month Standardized Precipitation Index",
        source=ECAD_ATBD,
        short_name="SPI6",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.TAS],
        qualifiers=[REFERENCE_PERIOD_INDEX],
        output_unit="",
    )
    SPI3 = StandardIndex(
        reference=ECAD_REFERENCE,
        indicator=XCLIM_BINDING.StandardizedPrecipitationIndex3(),
        definition="3-Month Standardized Precipitation Index",
        source=ECAD_ATBD,
        short_name="SPI3",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.TAS],
        qualifiers=[REFERENCE_PERIOD_INDEX],
        output_unit="",
    )



from __future__ import annotations

from datetime import datetime
from typing import Hashable, Sequence

import numpy as np
import xarray as xr
import xclim
from pint import Quantity
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset
from xclim.core.units import convert_units_to
from xclim.core.utils import PercentileDataArray

from icclim.generic_indices.standard_variable import (
    StandardVariable,
    StandardVariableRegistry,
)
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.icclim_types import InFileBaseType
from icclim.models.cf_calendar import CfCalendarRegistry
from icclim.models.constants import UNITS_KEY, VALID_PERCENTILE_DIMENSION
from icclim.models.standard_index import StandardIndex
from icclim.utils import get_date_to_iso_format, is_precipitation_amount

DEFAULT_INPUT_FREQUENCY = "days"


def guess_var_names(
    ds: Dataset,
    var_names: str | Sequence[str] | None,
    standard_index: StandardIndex | None,
) -> list[Hashable]:
    if var_names is None:
        return _guess_dataset_var_names(ds=ds, standard_index=standard_index)
    elif isinstance(var_names, str):
        return [var_names]
    elif isinstance(var_names, (list, tuple)):
        return var_names
    else:
        raise NotImplementedError("`var_name` must be a string a list or None.")


def read_dataset(
    in_files: InFileBaseType,
    standard_var: StandardVariable | None = None,
    var_name: str | Sequence[str] = None,
) -> Dataset:
    if isinstance(in_files, Dataset):
        ds = in_files
    elif isinstance(in_files, DataArray):
        ds = _read_dataarray(in_files, standard_var=standard_var, var_name=var_name)
    elif is_glob_path(in_files) or (
        isinstance(in_files, (list, tuple)) and is_netcdf_path(in_files[0])
    ):
        # we assumes it's a list of netCDF files
        #  join="override" is used for cases some dimension are a tiny bit different
        #  in different files (was the case with eobs).
        ds = xr.open_mfdataset(in_files, parallel=True, join="override")  # noqa
    elif is_netcdf_path(in_files):
        ds = xr.open_dataset(in_files)
    elif is_zarr_path(in_files):
        ds = xr.open_zarr(in_files)
    elif isinstance(in_files, (list, tuple)):
        return xr.merge(
            [
                read_dataset(in_file, standard_var, var_name[i])
                for i, in_file in enumerate(in_files)
            ]
        )
    else:
        raise NotImplementedError(
            f"`in_files` format {type(in_files)} was not" f" recognized."
        )
    return update_to_standard_coords(ds)


def update_to_standard_coords(ds: Dataset) -> Dataset:
    """
    Mutate input ds to use more icclim friendly coordinate names.
    """
    # TODO see if cf-xarray could replace this
    if ds.coords.get("t") is not None:
        ds = ds.rename({"t": "time"})
    return ds


def is_zarr_path(path: InFileBaseType) -> bool:
    return isinstance(path, str) and ".zarr" in path


def is_netcdf_path(path: InFileBaseType) -> bool:
    return isinstance(path, str) and ".nc" in path


def is_glob_path(path: InFileBaseType) -> bool:
    return isinstance(path, str) and "*" in path


def standardize_percentile_dim_name(per_da: DataArray) -> DataArray:
    # todo [xclim backport] This function could probably be in PercentileDataArray
    per_dim_name = None
    for d in VALID_PERCENTILE_DIMENSION:
        if d in per_da.dims:
            per_dim_name = d
        elif f"{d}s" in per_da.dims:
            # plural handling
            per_dim_name = f"{d}s"
    if per_dim_name is None:
        raise InvalidIcclimArgumentError(
            "Percentile data must contain a recognizable percentiles dimension such as"
            " 'percentiles', 'quantile', 'per' or 'centile'."
        )
    per_da = per_da.rename({per_dim_name: "percentiles"})
    if "quantile" in per_dim_name:
        per_da.coords["percentiles"] = per_da.coords["percentiles"] * 100
    return per_da


def read_clim_bounds(
    climatology_bounds: Sequence[str, str] | None, per_da: DataArray
) -> list[str]:
    bds = climatology_bounds or per_da.attrs.get("climatology_bounds", None)
    if len(bds) != 2:
        raise InvalidIcclimArgumentError(
            "climatology_bounds must be a iterable of length 2."
        )
    return [d for d in map(lambda bd: get_date_to_iso_format(bd), bds)]


def _read_dataarray(
    data: DataArray,
    standard_var: StandardVariable | None = None,
    var_name: str | Sequence[str] = None,
) -> Dataset:
    if isinstance(var_name, (tuple, list)):
        if len(var_name) > 1:
            raise InvalidIcclimArgumentError(
                "When the `in_file` is a DataArray, there"
                f" can only be one value in `var_name` but var_name was: {var_name} "
            )
        else:
            var_name = var_name[0]
        data_name = var_name or standard_var.short_name or None
    else:
        data_name = var_name or data.name or "unnamed_var"
    return data.to_dataset(name=data_name, promote_attrs=True)


def _guess_dataset_var_names(
    standard_index: StandardIndex, ds: Dataset
) -> list[Hashable]:
    """Try to guess the variable names using the expected kind of variable for
    the index.
    """
    if standard_index is not None:
        main_aliases = ", ".join(
            map(lambda v: v.short_name, standard_index.input_variables)
        )
        error_msg = (
            f"Index {standard_index.short_name} needs the following variable(s)"
            f" [{main_aliases}], but the input variables were {list(ds.data_vars)}."
            f" Use `var_name` parameter to explicitly set variable names."
        )
        if len(ds.data_vars) == 1:
            if len(standard_index.input_variables) != 1:
                raise InvalidIcclimArgumentError(error_msg)
            return [get_name_of_first_var(ds)]
        climate_var_names = []
        for expected_standard_var in standard_index.input_variables:
            for alias in expected_standard_var.aliases:
                # check if dataset contains this alias
                if _is_alias_valid(ds, alias):
                    climate_var_names.append(_get_actual_name(ds, alias))
                    break
        if len(climate_var_names) < len(standard_index.input_variables):
            raise InvalidIcclimArgumentError(error_msg)
        return climate_var_names
    else:
        if len(ds.data_vars) == 1:
            return [get_name_of_first_var(ds)]
        else:
            return find_standard_vars(ds)


def find_standard_vars(ds: Dataset) -> list[Hashable]:
    return [
        v
        for v in ds.data_vars
        if StandardVariableRegistry.lookup(str(v), no_error=True) is not None
    ]


def guess_input_type(data: DataArray) -> StandardVariable | None:
    cf_input = StandardVariableRegistry.lookup(str(data.name), no_error=True)
    if cf_input is None and data.attrs.get("standard_name", None) is not None:
        cf_input = StandardVariableRegistry.lookup(
            data.attrs.get("standard_name"), no_error=True
        )
    if cf_input is None:
        return None
    return cf_input


def build_studied_data(
    original_da: DataArray,
    time_range: Sequence[str] | None,
    ignore_Feb29th: bool,
    standard_var: StandardVariable | None,
) -> DataArray:
    if time_range is not None:
        check_time_range_pre_validity("time_range", time_range)
        time_range = [get_date_to_iso_format(x) for x in time_range]
        da = original_da.sel(time=slice(time_range[0], time_range[1]))
        check_time_range_post_validity(da, original_da, "time_range", time_range)
        if len(da.time) == 0:
            raise InvalidIcclimArgumentError(
                f"The given `time_range` {time_range} is out of the dataset time"
                f" period: {original_da.time.min().dt.floor('D').values}"
                f" - {original_da.time.max().dt.floor('D').values}."
            )
    else:
        da = original_da
    if ignore_Feb29th:
        da = xclim.core.calendar.convert_calendar(da, CfCalendarRegistry.NO_LEAP.name)
    if da.attrs.get(UNITS_KEY, None) is None and standard_var is not None:
        da.attrs[UNITS_KEY] = standard_var.default_units
    if is_precipitation_amount(da):
        da = xclim.core.units.amount2rate(da)
    da = da.chunk("auto")
    return da


def check_time_range_pre_validity(key: str, tr: Sequence[datetime | str]) -> None:
    if len(tr) != 2:
        raise InvalidIcclimArgumentError(
            f"The given `{key}` {tr}"
            f" has {len(tr)} elements."
            f" It must have exactly 2 dates."
        )


def check_time_range_post_validity(da, original_da, key: str, tr: list) -> None:
    if len(da.time) == 0:
        raise InvalidIcclimArgumentError(
            f"The given `{key}` {tr} is out of the sample time bounds:"
            f" {original_da.time.min().dt.floor('D').values}"
            f" - {original_da.time.max().dt.floor('D').values}."
        )


def _is_alias_valid(ds, alias) -> bool:
    for ds_var in ds.data_vars:
        if str(ds_var).upper() == alias.upper():
            return True
    return False


def _get_actual_name(ds, alias) -> str:
    for ds_var in ds.data_vars:
        if str(ds_var).upper() == alias.upper():
            return str(ds_var)
    raise KeyError(f"Could not find {alias} in dataset.")


def get_name_of_first_var(ds: Dataset) -> str:
    return str(ds.data_vars[list(ds.data_vars.keys())[0]].name)


def is_dataset_path(query: Sequence | str) -> bool:
    if isinstance(query, (tuple, list)):
        return all(map(lambda q: is_netcdf_path(q), query))
    return is_zarr_path(query) or is_glob_path(query) or is_netcdf_path(query)


def reduce_only_leap_years(da: DataArray) -> DataArray:
    reduced_list = []
    for _, val in da.groupby(da.time.dt.year):
        if val.time.dt.dayofyear.max() == 366:
            reduced_list.append(val)
    if not reduced_list:
        raise InvalidIcclimArgumentError(
            "No leap year in current dataset. Do not use `only_leap_years` parameter."
        )
    return xr.concat(reduced_list, "time")


def read_threshold_DataArray(
    thresh_da: DataArray,
    threshold_min_value: str | float,
    climatology_bounds: Sequence[str],
    unit: str,
):
    if PercentileDataArray.is_compatible(thresh_da):
        built_value = PercentileDataArray.from_da(
            standardize_percentile_dim_name(thresh_da),
            read_clim_bounds(climatology_bounds, thresh_da),
        )
        built_value.attrs["unit"] = unit
    else:
        if threshold_min_value:
            if isinstance(threshold_min_value, str):
                threshold_min_value = convert_units_to(threshold_min_value, thresh_da)
            # todo in prcptot the replacing value (np.nan) needs to be 0
            built_value = thresh_da.where(thresh_da > threshold_min_value, np.nan)
        else:
            built_value = thresh_da
        built_value.attrs["unit"] = unit
    return built_value


def build_reference_da(
    original_da: DataArray,
    base_period_time_range: Sequence[datetime | str] | None,
    only_leap_years: bool,
    percentile_min_value: Quantity | None,
) -> DataArray:
    reference = original_da
    if base_period_time_range:
        check_time_range_pre_validity("base_period_time_range", base_period_time_range)
        base_period_time_range = [
            get_date_to_iso_format(x) for x in base_period_time_range
        ]
        reference = original_da.sel(
            time=slice(base_period_time_range[0], base_period_time_range[1])
        )
        check_time_range_post_validity(
            reference, original_da, "base_period_time_range", base_period_time_range
        )
    if only_leap_years:
        reference = reduce_only_leap_years(original_da)
    if percentile_min_value is not None:
        percentile_min_value = convert_units_to(str(percentile_min_value), reference)
        reference = reference.where(reference >= percentile_min_value, np.nan)
    return reference



from __future__ import annotations

import dataclasses
from typing import Callable

import numpy as np
import xarray as xr
from xarray import DataArray

from icclim.models.registry import Registry


@dataclasses.dataclass
class CfCalendar:
    aliases: list[str]
    is_leap: Callable[[DataArray], np.ndarray]

    @property
    def name(self) -> str:
        return self.aliases[0]


# todo: the whole class might be useless with the latest cftime
#      (we don't need our own CfCalendar if we can do `da.time.dt.is_leap_year`)
class CfCalendarRegistry(Registry[CfCalendar]):
    """
    Calendars known in CF plus some additional custom aliases for convenience.
    The first value of the aliases is the calendar "main" name.

    aliases: List[str]
        All the possible aliases or poorly typed calendar names targeting the same
        calendar.
    is_leap: Callable
        It expects a DataArray argument of years such as `da.time.dt.year`.
        Returns a mask of the input telling if the value is part of a leap year or not.

    """

    _item_class = CfCalendar

    NO_LEAP = CfCalendar(
        ["noleap", "no_leap", "days_365", "days365", "365_day", "365day"],
        lambda da: np.full_like(da.shape, False, dtype=bool),
    )
    DAYS_360 = CfCalendar(
        ["360_day", "days_360", "360day", "days360"],
        lambda da: np.full_like(da.shape, False, dtype=bool),
    )
    ALL_LEAP = CfCalendar(
        ["all_leap", "allleap", "days_366", "days366", "366_day", "366day"],
        lambda da: np.full_like(da.shape, True, dtype=bool),
    )
    PROLEPTIC_GREGORIAN = CfCalendar(
        ["proleptic_gregorian", "prolepticgregorian"],
        lambda da: _proleptic_gregorian_leap(da).values,
    )
    JULIAN = CfCalendar(["julian"], lambda da: _julian_leap(da).values)
    STANDARD = CfCalendar(
        ["standard", "gregorian"], lambda da: _standard_leap(da).values
    )
    # Not sure what to do with none calendar
    NONE = CfCalendar(["none"], lambda da: _standard_leap(da).values)

    @staticmethod
    def get_item_aliases(item: CfCalendar) -> list[str]:
        return list(map(str.upper, item.aliases))


def _proleptic_gregorian_leap(years: DataArray) -> DataArray:
    return np.logical_or(
        years % 400 == 0, np.logical_and(years % 100 != 0, years % 4 == 0)
    )


def _julian_leap(years: DataArray) -> DataArray:
    return years % 4 == 0


def _standard_leap(years: DataArray) -> DataArray:
    res = xr.full_like(years, False)
    res[years < 1582] = _julian_leap(years[years < 1582])
    res[years >= 1582] = _proleptic_gregorian_leap(years[years >= 1582])
    return res



from __future__ import annotations

from abc import ABC
from copy import deepcopy
from typing import Generic, TypeVar

from icclim.icclim_exceptions import InvalidIcclimArgumentError

T = TypeVar("T")


class Registry(Generic[T], ABC):
    """This class is a fancy enum to easily store and find constant items of
    similar type.
    It acts as a namespace so there is no need to instantiate it or any of
    its subclasses.

    Notes
    -----
    Registries are not meant to store large collections, they are just fancy lookup
    tables for items with aliases and no case sensitivity.
    """

    _item_class: type  # runtime type for the generic `T`

    @classmethod
    def lookup(cls, query: T | str, no_error: bool = False) -> T:
        if isinstance(query, cls._item_class):
            return query
        if isinstance(query, str):
            q = query.upper()
            for key, item in cls.catalog().items():
                if q == key.upper() or q in cls.get_item_aliases(item):
                    return deepcopy(item)
        if no_error:
            return None
        raise InvalidIcclimArgumentError(
            f"Unknown {cls._item_class.__qualname__}: '{query}'. "
            f"Use one of {cls.every_aliases()}."
        )

    @classmethod
    def every_aliases(cls) -> list[T]:
        return list(map(cls.get_item_aliases, list(cls.catalog().values())))

    @staticmethod
    def get_item_aliases(item: T) -> list[str]:
        """Should be overridden."""
        return [item.name.upper()]

    @classmethod
    def catalog(cls) -> dict[str, T]:
        return {k: v for k, v in cls.__dict__.items() if isinstance(v, cls._item_class)}

    @classmethod
    def values(cls) -> list[T]:
        return [v for k, v in cls.__dict__.items() if isinstance(v, cls._item_class)]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))
import icclim

# -- Project information -----------------------------------------------------
project = "icclim"
copyright = "2021, CERFACS"
author = "Christian P."
version = icclim.__version__
release = icclim.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx_codeautolink",
    "sphinx_lfs_content",
    "sphinx_copybutton",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]

html_favicon = "_static/logo_icclim_favicon__displayed.ico"
html_theme_options = {
    "logo": {
        "image_light": "logo_icclim_colored__displayed.svg",
        "image_dark": "logo_icclim_white__displayed.svg",
    },
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/cerfacs-globc/icclim",  # required
            "icon": "fab fa-github-square",
            "type": "fontawesome",
        },
    ],
}

intersphinx_mapping = {
    "clisops": ("https://clisops.readthedocs.io/en/latest/", None),
    "flox": ("https://flox.readthedocs.io/en/latest/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "statsmodels": ("https://www.statsmodels.org/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "xclim": ("https://xclim.readthedocs.io/en/stable/", None),
}

# The master toctree document.
master_doc = "index"



from __future__ import annotations

from typing import TypedDict


class GlobalMetadata(TypedDict):
    history: str | None
    source: str | None
    time_encoding: dict | None  # to be read from ds.time.encoding
